package top.octopusyan.utils;

import javafx.application.Platform;
import javafx.scene.control.TextArea;
import org.apache.commons.io.FileUtils;
import top.octopusyan.manager.ThreadPoolManager;
import top.octopusyan.model.ProxySetupModel;

import java.io.*;
import java.nio.channels.FileChannel;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;

/**
 * @author : octopus yan
 * @email : octopus_yan@foxmail.com
 * @description : TODO frp 工具
 * @create : 2022-4-7 23:19
 */
public class FrpUtil {
    private static final String YAN_FRP_TEMP_DIR_NAME = "frpclienttmp";
    private static final String FRPC_CLIENT_FILE_NAME = "frpclienttmpfile.exe";
    private final String FRPC_CONF_PREFIX_NAME = "proxy_";
    private final String FRPC_CONF_SUFFIX_NAME = ".ini";
    private static final String YAN_FRP_TEMP_DIR_PATH = System.getProperty("java.io.tmpdir") + System.currentTimeMillis() + YAN_FRP_TEMP_DIR_NAME;
    private static final String FRPC_CLIENT_FILE_PATH = YAN_FRP_TEMP_DIR_PATH + File.separator + FRPC_CLIENT_FILE_NAME;
    private static final File frpcDir = new File(YAN_FRP_TEMP_DIR_PATH);
    public static final File frpc = new File(FRPC_CLIENT_FILE_PATH);
    public File frpcConfigFile;

    private final ProxySetupModel model;
    private final TextArea console;
    private Thread thread;
    private Process exec;


    private FrpUtil(TextArea console, ProxySetupModel model) {
        this.console = console;
        this.model = model;
    }

    public static FrpUtil init(ProxySetupModel model) {
        TextArea console = new TextArea();
        console.setPrefHeight(180);
        console.setMinWidth(200);
        return new FrpUtil(console, model);
    }

    /**
     * 初始化frpc客户端文件
     */
    public static void initFrpc() throws IOException {

        // 存在就删除
        if (frpcDir.exists()) FrpUtil.clearTmp();

        // 重新创建
        frpcDir.mkdir();
        frpc.createNewFile();

        // 发送删除请求
        frpcDir.deleteOnExit();
        frpc.deleteOnExit();

        // 复制 frpc
        String fprcFilePath = FrpUtil.class.getResource("/static/frpc.exe").getFile();
        System.out.println(fprcFilePath);
        copyFileUsingFileStreams(FrpUtil.class.getResourceAsStream("/static/frpc.exe"), frpc);
//        FileUtils.copyFile(new File(fprcFilePath), frpc);
    }

    /**
     * 清理缓存
     */
    public static void clearTmp() {
        try {
            FileUtils.deleteDirectory(frpcDir);
        } catch (IOException e) {
            e.printStackTrace();
            try {
                Runtime.getRuntime().exec("taskkill /im " + FRPC_CLIENT_FILE_NAME + " /f");
                Thread.sleep(2000);
                FileUtils.deleteDirectory(frpcDir);
            } catch (InterruptedException | IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    public TextArea getConsole() {
        return console;
    }

    /**
     * 开始
     */
    public void start() {

        thread = new Thread(() -> {

            Platform.runLater(() -> console.appendText("yan-frp-info:正在启动\n"));
            try {
                // 检查客户端文件
                if (!frpc.exists()) initFrpc();

                // 初始化配置文件
                initProxyConfigFile();

                // 执行命令
                String command = FRPC_CLIENT_FILE_PATH + " -c " + getConfigFilePath();
                System.out.println(command);
                exec = Runtime.getRuntime().exec(command, null, frpcDir);

                // 设置运行状态
                Platform.runLater(() -> model.setRunning(true));

                // 获取标准输出
                BufferedReader readStdout = new BufferedReader(new InputStreamReader(exec.getInputStream()));

                // 逐行读取
                String line;
                while ((line = readStdout.readLine()) != null) {
                    String finalLine = line;
                    Platform.runLater(() -> console.appendText("yan-frp-info:" + finalLine + "\n"));
                }

            } catch (IOException e) {
                e.printStackTrace();
                //
                Platform.runLater(() -> console.appendText("yan-frp-error:" + e.getMessage() + "\n"));
                Platform.runLater(() -> console.appendText("yan-frp-error:启动失败\n"));
            }
        });

        thread.start();
    }

    /**
     * 停止隧道
     */
    public void stop() {
        ThreadPoolManager.getInstance().execute(() -> {
            if (thread != null) {
                do {
                    if (exec != null && exec.isAlive())
                        exec.destroy();

                    thread.interrupt();

                } while (!thread.isInterrupted());

                Platform.runLater(() -> {
                    model.setRunning(false);
                    console.appendText("yan-frp-info:已停止\n");
                });

                // 尝试删除缓存配置文件
                if (frpcConfigFile.exists()) frpcConfigFile.delete();
            }
        });
    }

    /**
     * 清理日志
     */
    public void clearLog() {
        if (console != null)
            Platform.runLater(() -> console.clear());
    }

    /**
     * 隧道配置文件地址
     */
    private String getConfigFilePath() {
        return YAN_FRP_TEMP_DIR_PATH + File.separator + FRPC_CONF_PREFIX_NAME + model.getProxyName() + FRPC_CONF_SUFFIX_NAME;
    }

    /**
     * 初始化隧道配置文件
     */
    private void initProxyConfigFile() throws IOException {
        // 创建配置文件
        frpcConfigFile = new File(getConfigFilePath());
        if (frpcConfigFile.exists()) {
            frpcConfigFile.delete();
        }
        frpcConfigFile.createNewFile();
        frpcConfigFile.deleteOnExit();

        // 写入服务配置
        FileUtils.write(frpcConfigFile, ProxyUtil.getUserServerConfig(model.get().getNode()), StandardCharsets.UTF_8);
        // 写入隧道配置
        FileUtils.write(frpcConfigFile, ProxyUtil.getProxyConfig(model), StandardCharsets.UTF_8, true);
    }


    private static void copyFileUsingFileStreams(InputStream input, File dest)
            throws IOException {
        OutputStream output = null;
        try {
            output = new FileOutputStream(dest);
            byte[] buf = new byte[1024];
            int bytesRead;
            while ((bytesRead = input.read(buf)) > 0) {
                output.write(buf, 0, bytesRead);
            }
        } finally {
            input.close();
            output.close();
        }
    }
}
